import emerge as em
import numpy as np
import matplotlib.pyplot as plt

""" BANDPASS FILTER SYNTHESIS DEMO

This demo synthesizes a 7-section rectangular waveguide bandpass filter
using iris inverters. First, iris scattering parameters are simulated to
extract K-inverter and phase data. Then the full filter geometry is
assembled and S-parameters are plotted. """


# --- Physical constants and units ---------------------------------------
mm = 0.001                # meters per millimeter
dielectric_constant = 1.0 # (not used for air-filled waveguide)
c0 = 299792458            # speed of light in vacuum (m/s)

# --- Waveguide and frequency definitions --------------------------------
wga = 22.86 * mm          # a-dimension of WR-90 waveguide
wgb = 10.16 * mm              # b-dimension (height) of waveguide
# frequency band for final filter response
f1 = 9.6e9                # start frequency (Hz)
f2 = 10e9                # stop frequency (Hz)

# feed line length and iris thickness
Lfeed = 20 * mm           # length of input/output waveguide sections
t_thickness = 1 * mm      # iris plate thickness

# Prototype filter g-coefficients for 7 sections (n+2 values)
gs = [1.000, 1.1811, 1.4228, 2.0966, 1.5733, 2.0966, 1.4228, 1.1811, 1.000]

# --- Helper functions for wave propagation and inverter theory ---------
# Calculate axial propagation constant for TE10 mode at frequency f
def kz(f: float):
    beta = 2 * np.pi * f / c0
    k_c = np.pi / wga       # cutoff wavenumber for TE10
    return np.sqrt(beta**2 - k_c**2)

# Center frequency and fractional bandwidth
f0 = np.sqrt(f1 * f2)
beta0 = kz(f0)
# Fractional electrical length difference between band edges
delta = (2*np.pi/kz(f1) - 2*np.pi/kz(f2)) / (2*np.pi/kz(f0))

# Number of inverter sections (exclude terminations)
N = len(gs) - 2
Kvals = []  # to store theoretical inverter values

# Compute K-inverter coefficients for prototype
for i, (g_prev, g_next) in enumerate(zip(gs[:-1], gs[1:])):
    if i == 0 or i == N:
        # end inverters use square-root formula
        Kvals.append(np.sqrt(np.pi * delta / (2 * g_prev * g_next)))
    else:
        # interior inverters use linear formula
        Kvals.append(np.pi * delta / (2 * np.sqrt(g_prev * g_next)))

# Functions to extract susceptances from scattering parameters
def XsZ0(S11, S12, S21, S22):
    # series susceptance (normalized)
    return np.real(((1 - S12 + S11) / (1 - S11 + S12)) / 1j)

def XpZ0(S11, S12, S21, S22):
    # parallel susceptance (normalized)
    return np.real((2 * S12 / ((1 - S11)**2 - S12**2)) / 1j)

def phif(S11, S12, S21, S22):
    # electrical phase shift of inverter
    xs = XsZ0(S11, S12, S21, S22)
    xp = XpZ0(S11, S12, S21, S22)
    return -np.arctan(2*xp + xs) - np.arctan(xs)

def KZ0(S11, S12, S21, S22):
    # compute actual K/Z0 from phase and susceptance
    return np.abs(np.tan(phif(S11,S12,S21,S22)/2 + np.arctan(XsZ0(S11,S12,S21,S22))))

# --- Iris parameter sweep to characterize inverter -----------------------
# Sweep iris gap from 1 mm to 20 mm and compute K and phi at f0
wgaps = np.linspace(1*mm, 20*mm, 21)
Ks = []
hphis = []
with em.Simulation('IrisSim') as sim:
    sim.check_version("0.6.8") # Checks version compatibility.
    for wgap in sim.parameter_sweep(True, wgap=wgaps):
        # Define two short waveguide sections separated by iris plate
        wg1 = em.geo.Box(wga, Lfeed, wgb, (-wga/2, -Lfeed - t_thickness/2, 0))
        iris = em.geo.Box(wgap, t_thickness, wgb, (-wgap/2, -t_thickness/2, 0))
        wg2 = em.geo.Box(wga, Lfeed, wgb, (-wga/2, t_thickness/2, 0))

        sim.commit_geometry()
        sim.mw.set_frequency(f0)
        sim.mw.set_resolution(0.10)
        sim.mesher.set_domain_size(iris, 2*mm)
        sim.generate_mesh()

        # Port BCs on front/back faces for TE10 excitation
        sim.mw.bc.RectangularWaveguide(wg1.face('front'), 1)
        sim.mw.bc.RectangularWaveguide(wg2.face('back'), 2)

        data = sim.mw.run_sweep()
        # Compensate phase from feed line length
        S11 = data.scalar.select(wgap=wgap).S(1,1) * np.exp(1j*2*kz(f0)*Lfeed)
        S12 = data.scalar.select(wgap=wgap).S(1,2) * np.exp(1j*2*kz(f0)*Lfeed)
        S21 = data.scalar.select(wgap=wgap).S(2,1) * np.exp(1j*2*kz(f0)*Lfeed)
        S22 = data.scalar.select(wgap=wgap).S(2,2) * np.exp(1j*2*kz(f0)*Lfeed)

        Ks.append(KZ0(S11, S12, S21, S22))
        hphis.append(phif(S11, S12, S21, S22))

# --- Plot the K-inverter graph ------------------------------------------

fig, ax = plt.subplots()
ax.plot(wgaps*1000, Ks)
ax.grid(True)
ax.set_xlabel('Iris gap [mm]')
ax.set_ylabel('K/Z0')
plt.show()

# --- Interpolate iris widths and electrical lengths ---------------------
Ks = np.array(Ks)
# Map theoretical Kvals to physical iris gaps
interp_widths = [np.interp(K, Ks, wgaps) for K in Kvals]
# Map theoretical phase shifts to physical iris gaps
interp_phis = [np.interp(K, Ks, np.array(hphis)) for K in Kvals]
# Compute required cavity lengths from phase delays
cavity_lengths = (1/beta0 * np.array([
    np.pi + 0.5*(interp_phis[i] + interp_phis[i+1])
    for i in range(len(Kvals)-1)
])).real

# --- Build and simulate full filter -------------------------------------
with em.Simulation('FullFilter') as mf:
    # Input feed section
    feed1 = em.geo.Box(wga, Lfeed, wgb, (-wga/2, -Lfeed, 0))
    # Create cavities and irises sequentially
    y0 = 0
    cavities = []
    irises = []
    for L, W in zip(cavity_lengths, interp_widths[:-1]):
        ir = em.geo.Box(W, t_thickness, wgb, (-W/2, y0, 0))
        cav = em.geo.Box(wga, L, wgb, (-wga/2, y0 + t_thickness, 0))
        y0 += L + t_thickness
        irises.append(ir)
        cavities.append(cav)
    # Last iris and output feed
    last_iris = em.geo.Box(interp_widths[-1], t_thickness, wgb, (-interp_widths[-1]/2, y0, 0))
    feed2 = em.geo.Box(wga, Lfeed, wgb, (-wga/2, y0 + t_thickness, 0))

    # Define the full filter geometry
    mf.commit_geometry(feed1, feed2, last_iris, *(cavities + irises))

    # Simulation settings and mesh
    mf.mw.set_frequency_range(f1 - 0.2e9, f2 + 0.2e9, 51)
    mf.mw.set_resolution(0.10)
    for ir in irises:
        mf.mesher.set_domain_size(ir, 2*mm)
    mf.generate_mesh()

    # Boundary conditions for feed ports
    p1 = mf.mw.bc.RectangularWaveguide(feed1.face('front'), 1)
    p2 = mf.mw.bc.RectangularWaveguide(feed2.face('back'), 2)

    # Run frequency-domain sweep and extract S-parameters
    data = mf.mw.run_sweep(parallel=True, njobs=3, frequency_groups=8)
    grid = data.scalar.grid
    freqs = grid.freq
    fdense = np.linspace(freqs[0], freqs[-1], 2001)

    S11 = grid.model_S(1,1,fdense)
    S21 = grid.model_S(2,1,fdense)

    # Plot the filter response (dB)
    fig, ax = plt.subplots()
    ax.plot(fdense, 20*np.log10(np.abs(S11)), label='S11')
    ax.plot(fdense, 20*np.log10(np.abs(S21)), label='S21')
    ax.set_xlabel('Frequency [GHz]')
    ax.set_ylabel('S-parameter [dB]')
    ax.grid(True)
    ax.legend()
    plt.show()

    # Visualize geometry and mode shapes
    mf.display.add_object(feed1, opacity=0.1)
    mf.display.add_object(feed2, opacity=0.1)
    for obj in irises + cavities:
        mf.display.add_object(obj, opacity=0.1)
    # Show electric field cut-plane at center frequency
    cut = data.field.find(freq=f0).cutplane(2*mm, z=wgb/2)
    mf.display.add_surf(*cut.scalar('Ez','real'))
    mf.display.add_portmode(p1)
    mf.display.add_portmode(p2)
    mf.display.show()



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.


"""
THESE FUNCTIONS ARE MEANT TO QUICKLY COUPLE COORDINATES ON TWO PERIODIC
BOUNDARIES IN THE SMALLEST AMOUNT OF TIME WITHOUT RISKING FLOAT ROUNDING ERRORS.

THE MAPPING IS DONE BY BY SIMPLY FINDING THE CLOSEST COORDINATE AFTER PROJECTION
BY SEARCHING A LIST SORTED DICTIONARY WISE ON 3D COORDINATES.

ALL PAIRED NODES ARE NO LONGER VISITED
"""

import numpy as np
from numba import njit, f8, i8 # type: ignore


############################################################
#                  NUMBA COMPILED FUNCTION                 #
############################################################


@njit(i8[:,:](f8[:,:], i8[:], i8[:], f8[:], f8), cache=True, nogil=True)
def link_coords(coords: np.ndarray, ids1: np.ndarray, ids2: np.ndarray, disp: np.ndarray, dsmax: float) -> np.ndarray:
    D = dsmax**2
    N = ids1.shape[0]
    ids2_mapped = np.zeros_like(ids2)

    available = np.ones_like(ids2)
    id_start = 0

    for i1 in range(N):
        ictr = 0
        c1 = coords[:,ids1[i1]]
        for i2 in range(id_start, N):
            if available[i2] == 0:
                continue
            c2 = coords[:,ids2[i2]]-disp
            dist = (c2[0]-c1[0])**2 + (c2[1]-c1[1])**2 + (c2[2]-c1[2])**2
            if dist > D:
                ictr += 1
                continue
            if ictr==0:
                id_start += 1
            ids2_mapped[i1] = ids2[i2]
            available[i2] = 0
            break
        
    out = np.zeros((2, N), dtype=np.int64)
    out[0,:] = ids1
    out[1,:] = ids2_mapped
    return out


############################################################
#                   MAIN PYTHON INTERFACE                  #
############################################################

def pair_coordinates(coords: np.ndarray, ids1: np.ndarray, ids2: np.ndarray, disp: np.ndarray, dsmax: float) -> dict[int, int]:
    """ This function finds the mapping between a total coordinate set and two lits of indices.

    The indices correspond to two faces that are identical but displaced (mesh centroids of periodic boundaries).

    Args:
        coords (np.ndarray): A total set of coordinates
        ids1 (list[int]): The indices of the source set
        ids2 (list[int]): The indices of the to-be-matched set
        disp (np.ndarray): The displacement vector of shape (3,)
        dsmax (float): The maximum allowed displacement in matchiing

    Returns:
        dict[int, int]: An int,int mapping of the indices.
    """
    ids1_c_sorted = sorted(ids1, key= lambda x: tuple(coords[:,x]))
    ids2_c_sorted = sorted(ids2, key= lambda x: tuple(coords[:,x]-disp))

    mapping = link_coords(coords, np.array(ids1_c_sorted), np.array(ids2_c_sorted), disp, dsmax)
    
    mapping = {i: j for i,j in zip(mapping[0,:], mapping[1,:])}
    
    return mapping


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import gmsh # type: ignore
from .material import Material, AIR
from .selection import FaceSelection, DomainSelection, EdgeSelection, PointSelection, Selection
from loguru import logger
from typing import Literal, Any, Iterable, TypeVar
import numpy as np



def _map_tags(tags: list[int], mapping: dict[int, list[int]]):
    new_tags = []
    for tag in tags:
        new_tags.extend(mapping.get(tag, [tag,]))
    return new_tags

def _bbcenter(x1, y1, z1, x2, y2, z2):
    return np.array([(x1+x2)/2, (y1+y2)/2, (z1+z2)/2])
FaceNames = Literal['back','front','left','right','top','bottom']

class _KEY_GENERATOR:

    def __init__(self):
        self.start = -1
    
    def new(self) -> int:
        self.start += 1
        return self.start

class _GeometryManager:

    def __init__(self):
        self.geometry_list: dict[str, list[GeoObject]] = dict()
        self.active: str = ''

    def all_geometries(self, model: str | None = None) -> list[GeoObject]:
        if model is None:
            model = self.active
        return [geo for geo in self.geometry_list[model] if geo._exists]

    def submit_geometry(self, geo: GeoObject, model: str | None = None) -> None:
        if model is None:
            model = self.active
        self.geometry_list[model].append(geo)

    def sign_in(self, modelname: str) -> None:
        if modelname not in self.geometry_list:
            self.geometry_list[modelname] = []
        self.active = modelname

    def reset(self, modelname: str) -> None:
        self.geometry_list[modelname] = []
    
    def lowest_priority(self) -> int:
        return min([geo._priority for geo in self.all_geometries()])
    
    def highest_priority(self) -> int:
        return min([geo._priority for geo in self.all_geometries()])
    
class _FacePointer:
    """The FacePointer class defines a face to be selectable as a
    face normal vector plus an origin. All faces of an object
    can be selected based on the projected distance to the defined
    selection plane of the center of mass of a face iff the normals
    also align with some tolerance.

    """
    def __init__(self, 
                 origin: np.ndarray, 
                 normal: np.ndarray):
        self.o = np.array(origin)
        self.n = np.array(normal)

    def find(self, dimtags: list[tuple[int,int]],
             origins: list[np.ndarray],
             normals: list[np.ndarray]) -> list[int]:
        tags = []
        for (d,t), o, n in zip(dimtags, origins, normals):
            normdist = np.abs((o-self.o)@self.n)
            dotnorm = np.abs(n@self.n)
            if normdist < 1e-3 and dotnorm > 0.99:
                tags.append(t)
        return tags
    
    def rotate(self, c0, ax, angle):
        """
        Rotate self.o and self.n about axis `ax`, centered at `c0`, by `angle` radians.

        Parameters
        ----------
        c0 : np.ndarray
            The center of rotation, shape (3,).
        ax : np.ndarray
            The axis to rotate around, shape (3,). Need not be unit length.
        angle : float
            Rotation angle in radians.
        """
        angle = -angle
        # Ensure axis is a unit vector
        k = ax / np.linalg.norm(ax)

        # Precompute trig values
        cos_theta = np.cos(angle)
        sin_theta = np.sin(angle)

        def rodrigues(v: np.ndarray) -> np.ndarray:
            """
            Rotate vector v around axis k by angle using Rodrigues' formula.
            """
            # term1 = v * cosθ
            term1 = v * cos_theta
            # term2 = (k × v) * sinθ
            term2 = np.cross(k, v) * sin_theta
            # term3 = k * (k ⋅ v) * (1 - cosθ)
            term3 = k * (np.dot(k, v)) * (1 - cos_theta)
            return term1 + term2 + term3

        # Rotate the origin point about c0:
        rel_o = self.o - c0            # move to rotation-centre coordinates
        rot_o = rodrigues(rel_o)       # rotate
        self.o = rot_o + c0            # move back

        # Rotate the normal vector (pure direction, no translation)
        self.n = rodrigues(self.n)

    def translate(self, dx, dy, dz):
        self.o = self.o + np.array([dx, dy, dz])
    
    def mirror(self, c0: np.ndarray, pln: np.ndarray) -> None:
        """
        Reflect self.o and self.n across the plane passing through c0
        with normal pln.

        Parameters
        ----------
        c0 : np.ndarray
            A point on the mirror plane, shape (3,).
        pln : np.ndarray
            The normal of the mirror plane, shape (3,). Need not be unit length.
        """
        # Normalize the plane normal
        k = pln / np.linalg.norm(pln)
        

        # Reflect the origin point:
        # compute vector from plane point to self.o
        v_o = self.o - c0
        # signed distance along normal
        dist_o = np.dot(v_o, k)
        # reflection
        self.o = self.o - 2 * dist_o * k

        # Reflect the normal/direction vector:
        dist_n = np.dot(self.n, k)
        self.n = self.n - 2 * dist_n * k

    def affine_transform(self, M: np.ndarray):
        """
        Apply a 4×4 affine transformation matrix to both self.o and self.n.

        Parameters
        ----------
        M : np.ndarray
            The 4×4 affine transformation matrix.
            - When applied to a point, use homogeneous w=1.
            - When applied to a direction/vector, use homogeneous w=0.
        """
        # Validate shape
        if M.shape != (4, 4):
            raise ValueError(f"Expected M to be 4×4, got shape {M.shape}")

        # Transform origin point (homogeneous w=1)
        homo_o = np.empty(4)
        homo_o[:3] = self.o
        homo_o[3] = 1.0
        transformed_o = M @ homo_o
        self.o = transformed_o[:3]

        # Transform normal/direction vector (homogeneous w=0)
        homo_n = np.empty(4)
        homo_n[:3] = self.n
        homo_n[3] = 0.0
        transformed_n = M @ homo_n
        self.n = transformed_n[:3]
        # Optionally normalize self.n if you need to keep it unit-length:
        # self.n = self.n / np.linalg.norm(self.n)

    def copy(self) -> _FacePointer:
        return _FacePointer(self.o, self.n)


_GENERATOR = _KEY_GENERATOR()
_GEOMANAGER = _GeometryManager()

class GeoObject:
    """A generalization of any OpenCASCADE entity described by a dimension and a set of tags.
    """
    dim: int = -1
    def __init__(self, tags: list[int] | None = None):
        if tags is None:
            tags = []
        self.old_tags: list[int] = []
        self.tags: list[int] = tags
        self.material: Material = AIR
        self.mesh_multiplier: float = 1.0
        self.max_meshsize: float = 1e9
        
        self._unset_constraints: bool = False
        self._embeddings: list[GeoObject] = []
        self._face_pointers: dict[str, _FacePointer] = dict()
        self._tools: dict[int, dict[str, _FacePointer]] = dict()
        
        self._key = _GENERATOR.new()
        self._aux_data: dict[str, Any] = dict()
        self._priority: int = 10

        self._exists: bool = True
        _GEOMANAGER.submit_geometry(self)

    @property
    def color_rgb(self) -> tuple[float, float, float]:
        return self.material.color_rgb
    
    @property
    def opacity(self) -> float:
        return self.material.opacity
    
    @property
    def select(self) -> Selection:
        '''Returns a corresponding Face/Domain or Edge Selection object'''
        if self.dim==1:
            return EdgeSelection(self.tags)
        elif self.dim==2:
            return FaceSelection(self.tags)
        elif self.dim==3:
            return DomainSelection(self.tags)
        else:
            return Selection(self.tags)
    
    @staticmethod
    def merged(objects: list[GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject]) -> list[GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject] | GeoPoint | GeoEdge | GeoSurface | GeoVolume | GeoObject:
        dim = objects[0].dim
        tags = []
        out: GeoObject | None = None
        for obj in objects:
            tags.extend(obj.tags)
        if dim==2:
            out = GeoSurface(tags)
        elif dim==3:
            out = GeoVolume(tags)
        else:
            out = GeoObject(tags)
        out.material = objects[0].material
        return out
    
    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.dim},{self.tags})'

    def _data(self, *labels) -> tuple[Any | None, ...]:
        return tuple([self._aux_data[lab] for lab in labels])
    
    def _add_face_pointer(self, 
                          name: str,
                          origin: np.ndarray,
                          normal: np.ndarray):
        self._face_pointers[name] = _FacePointer(origin, normal)
    
    def make_copy(self) -> GeoObject:
        new_dimtags = gmsh.model.occ.copy(self.dimtags)
        new_obj = GeoObject.from_dimtags(new_dimtags)
        new_obj.material = self.material
        new_obj.mesh_multiplier = self.mesh_multiplier
        new_obj.max_meshsize = self.max_meshsize
        
        new_obj._unset_constraints = self._unset_constraints
        new_obj._embeddings = [emb.make_copy() for emb in self._embeddings]
        new_obj._face_pointers = {key: value.copy() for key,value in self._face_pointers.items()}
        new_obj._tools = {key: {key2: value2.copy() for key2, value2 in value.items()} for key,value in self._tools.items()}
        
        new_obj._aux_data = self._aux_data.copy()
        new_obj._priority = self._priority
        new_obj._exists = self._exists
        return new_obj

    def replace_tags(self, tagmap: dict[int, list[int]]):
        self.old_tags = self.tags
        newtags = []
        for tag in self.tags:
            newtags.extend(tagmap.get(tag, [tag,]))
        self.tags = newtags
        logger.debug(f'{self} Replaced {self.old_tags} -> {self.tags}')
    
    def update_tags(self, tag_mapping: dict[int,dict]) -> GeoObject:
        ''' Update the tag definition of a GeoObject after fragementation.'''
        self.replace_tags(tag_mapping[self.dim])
        return self
    
    def _take_pointers(self, *others: GeoObject) -> GeoObject:
        for other in others:
            self._face_pointers.update(other._face_pointers)
            self._tools.update(other._tools)
        return self
    
    @property
    def _all_pointers(self) -> list[_FacePointer]:
        pointers = list(self._face_pointers.values())
        for dct in self._tools.values():
            pointers.extend(list(dct.values()))
        return pointers
    
    @property
    def _all_pointer_names(self) -> set[str]:
        keys = set(self._face_pointers.keys())
        for dct in self._tools.values():
            keys = keys.union(set(dct.keys()))
        return keys
    
    def _take_tools(self, *objects: GeoObject) -> GeoObject:
        for obj in objects:
            self._tools[obj._key] = obj._face_pointers
            self._tools.update(obj._tools)
        return self
    
    def _face_tags(self, name: FaceNames, tool: GeoObject | None = None) -> list[int]:
        names = self._all_pointer_names
        if name not in names:
            raise ValueError(f'The face {name} does not exist in {self}')
        
        gmsh.model.occ.synchronize()
        dimtags = gmsh.model.get_boundary(self.dimtags, True, False)
        
        normals = [gmsh.model.get_normal(t, [0,0]) for d,t, in dimtags]
        origins = [gmsh.model.occ.get_center_of_mass(d, t) for d,t in dimtags]
        
        if tool is not None:
            tags = self._tools[tool._key][name].find(dimtags, origins, normals)
        else:
            tags = self._face_pointers[name].find(dimtags, origins, normals)
        logger.info(f'Selected face {tags}.')
        return tags

    def set_material(self, material: Material) -> GeoObject:
        self.material = material
        return self
    
    def prio_set(self, level: int) -> GeoObject:
        """Defines the material assignment priority level of this geometry.
        By default all objects have priority level 10. If you assign a lower number,
        in cases where multiple geometries occupy the same volume, the highest priority
        will be chosen.

        Args:
            level (int): The priority level

        Returns:
            GeoObject: The same object
        """
        self._priority = level
        return self
    
    def prio_up(self) -> GeoObject:
        """Increases the material selection priority by 1

        Returns:
            GeoObject: _description_
        """
        self._priority += 1
        return self
    
    def prio_down(self) -> GeoObject:
        """Decreases the material selection priority by 1

        Returns:
            GeoObject: _description_
        """
        self._priority -= 1
        return self

    def background(self) -> GeoObject:
        """Set the material selection priority to be on the background.

        Returns:
            GeoObject: _description_
        """
        self._priority = _GEOMANAGER.lowest_priority()-10
        return self

    def foreground(self) -> GeoObject:
        """Set the material selection priority to be on top.

        Returns:
            GeoObject: _description_
        """
        self._priority = _GEOMANAGER.highest_priority()+10
        return self
    
    def boundary(self, exclude: tuple[FaceNames,...] | None = None, tags: list[int] | None = None) -> FaceSelection:
        """Returns the complete set of boundary faces.

        If implemented, it is possible to exclude a set of faces based on their name
        or a list of tags (integers)

        Returns:
            FaceSelection: The selected faces
        """
        if exclude is None:
            exclude = tuple()
            
        if tags is None:
            tags = []
        
        
        for name in exclude:
            tags.extend(self.face(name).tags)
        dimtags = gmsh.model.get_boundary(self.dimtags, True, False)
        return FaceSelection([t for d,t in dimtags if t not in tags])
    
    def face(self, name: FaceNames, tool: GeoObject | None = None) -> FaceSelection:
        """Returns the FaceSelection for a given face name.
        
        The face name must be defined for the type of geometry.

        Args:
            name (FaceNames): The name of the face to select.

        Returns:
            FaceSelection: The selected face
        """
        
        return FaceSelection(self._face_tags(name, tool))

    def faces(self, *names: FaceNames, tool: GeoObject | None = None) -> FaceSelection:
        """Returns the FaceSelection for a given face names.
        
        The face name must be defined for the type of geometry.

        Args:
            name (FaceNames): The name of the face to select.

        Returns:
            FaceSelection: The selected face
        """
        tags = []
        for name in names:
            tags.extend(self._face_tags(name, tool))
        return FaceSelection(tags)
        
    @property
    def dimtags(self) -> list[tuple[int, int]]:
        return [(self.dim, tag) for tag in self.tags]
    
    @property
    def embeddings(self) -> list[tuple[int,int]]:
        return []

    @staticmethod
    def from_dimtags(dimtags: list[tuple[int,int]]) -> GeoVolume | GeoSurface | GeoObject:
        dim = dimtags[0][0]
        tags = [t for d,t in dimtags]
        if dim==0:
            return GeoPoint(tags)
        elif dim==1:
            return GeoEdge(tags)
        if dim==2:
            return GeoSurface(tags)
        if dim==3:
            return GeoVolume(tags)
        return GeoObject(tags)
    
    def remove(self) -> None:
        self._exists = False
        gmsh.model.occ.remove(self.dimtags, True)
        
class GeoVolume(GeoObject):
    '''GeoVolume is an interface to the GMSH CAD kernel. It does not represent EMerge
    specific geometry data.'''
    dim = 3
    def __init__(self, tag: int | Iterable[int]):
        super().__init__()
        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]

    @property
    def select(self) -> DomainSelection:
        return DomainSelection(self.tags)
    
class GeoPoint(GeoObject):
    dim = 0

    @property
    def select(self) -> PointSelection:
        return PointSelection(self.tags)
    
    def __init__(self, tag: int | list[int]):
        super().__init__()

        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]

class GeoEdge(GeoObject):
    dim = 1

    @property
    def select(self) -> EdgeSelection:
        return EdgeSelection(self.tags)
    
    def __init__(self, tag: int | list[int]):
        super().__init__()
        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]
        

class GeoSurface(GeoObject):
    '''GeoVolume is an interface to the GMSH CAD kernel. It does not reprsent Emerge
    specific geometry data.'''
    dim = 2

    @property
    def select(self) -> FaceSelection:
        return FaceSelection(self.tags)
    
    def __init__(self, tag: int | list[int]):
        super().__init__()
        self.tags: list[int] = []
        if isinstance(tag, Iterable):
            self.tags = list(tag)
        else:
            self.tags = [tag,]

class GeoPolygon(GeoSurface):
    
    def __init__(self,
                 tags: list[int]):
        super().__init__(tags)
        self.points: list[int] = []
        self.lines: list[int] = []


############################################################
#                      SHORT FUNCTIONS                     #
############################################################

def select(*items: Selection | GeoObject) -> Selection:
    """Generate a selection from a series of selections and/or objects that share the same dimension.

    Raises:
        ValueError: Raised if the dimensions provided are not consistent

    Returns:
        Selection: An output selection object.
    """
    dim = items[0].dim
    tags = []
    for item in items:
        if item.dim!=dim:
            raise ValueError(f'Cannot group of objects with a dissimilar dimensions. Trying to include {item} in a list of dimension {dim}.')
        tags.extend(item.tags)
    return Selection.from_dim_tags(dim, tags)
            



from ._emerge.solver import Solver, EigSolver, Preconditioner, Sorter
from ._emerge.geometry import GeoVolume, GeoSurface, GeoEdge, GeoPoint, _FacePointer
from ._emerge.physics.microwave.microwave_bc import RobinBC, PortBC


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from ..microwave_bc import PEC, BoundaryCondition, RectangularWaveguide, RobinBC, PortBC, Periodic, MWBoundaryConditionSet
from ....elements.nedelec2 import Nedelec2
from ....elements.nedleg2 import NedelecLegrange2
from ....mth.optimized import gaus_quad_tri
from ....mth.pairing import pair_coordinates
from ....material import Material
from scipy.sparse import csr_matrix
from loguru import logger
from ..simjob import SimJob
from .periodicbc import gen_periodic_matrix
from ....const import MU0, EPS0, C0


############################################################
#                         FUNCTIONS                        #
############################################################

def diagnose_matrix(mat: np.ndarray) -> None:

    if not isinstance(mat, np.ndarray):
        logger.debug('Converting sparse array to flattened array')
        mat = mat[mat.nonzero()].A1
        #mat = np.array(nonzero_mat)
    
    ''' Prints all indices of Nan's and infinities in a matrix '''
    ids = np.where(np.isnan(mat))
    if len(ids[0]) > 0:
        logger.error(f'Found NaN at {ids}')
    ids = np.where(np.isinf(mat))
    if len(ids[0]) > 0:
        logger.error(f'Found Inf at {ids}')
    ids = np.where(np.abs(mat) > 1e10)
    if len(ids[0]) > 0:
        logger.error(f'Found large values at {ids}')
    logger.info('Diagnostics finished')

def plane_basis_from_points(points: np.ndarray) -> np.ndarray:
    """
    Compute an orthonormal basis from a cloud of 3D points dominantly
    lying on one plane.

    Parameters
    ----------
    points : ndarray, shape (3, N)
        3D coordinates of the point cloud.

    Returns
    -------
    basis : ndarray, shape (3, 3)
        Matrix whose columns are:
            - first principal direction (plane X axis)
            - second principal direction (plane Y axis)
            - plane normal vector (Z axis)
    """
    if points.shape[0] != 3:
        raise ValueError("Input must have shape (3, N)")

    # Compute centroid
    centroid = points.mean(axis=1, keepdims=True)

    # Center the data
    points_centered = points - centroid

    # Compute covariance matrix (3x3)
    C = (points_centered @ points_centered.T) / points.shape[1]

    # Eigen decomposition
    eigvals, eigvecs = np.linalg.eigh(C)

    # Sort eigenvectors by descending eigenvalue
    idx = np.argsort(eigvals)[::-1]
    eigvecs = eigvecs[:, idx]

    # Columns of eigvecs = principal axes
    return eigvecs


############################################################
#                    THE ASSEMBLER CLASS                   #
############################################################

class Assembler:
    """The assembler class is responsible for FEM EM problem assembly.

    It stores some cached properties to accellerate preformance.
    """
    def __init__(self):
        
        self.cached_matrices = None
        self.conductivity_limit = 1e7
        # Currently not used.
        #self._Pmat_cache: dict[tuple[int,int], csr_matrix] = dict()
        #self._remove_cache: list[int] = []
    
    def assemble_bma_matrices(self,
                              field: Nedelec2,
                        er: np.ndarray, 
                        ur: np.ndarray, 
                        sig: np.ndarray,
                        k0: float,
                        port: PortBC,
                        bc_set: MWBoundaryConditionSet) -> tuple[csr_matrix, csr_matrix, np.ndarray, NedelecLegrange2]:
        """Computes the boundary mode analysis matrices

        Args:
            field (Nedelec2): The Nedelec2 field object
            er (np.ndarray): The relative permittivity tensor of shape (3,3,N)
            ur (np.ndarray): The relative permeability tensor of shape (3,3,N)
            sig (np.ndarray): The conductivity scalar of shape (N,)
            k0 (float): The simulation phase constant
            port (PortBC): The port boundary condition object
            bcs (MWBoundaryConditionSet): The other boundary conditions

        Returns:
            tuple[np.ndarray, np.ndarray, np.ndarray, NedelecLegrange2]: The E, B, solve ids and Mixed order field object.
        """
        from .generalized_eigen_hb import generelized_eigenvalue_matrix
        logger.debug('Assembling Boundary Mode Matrices')

        bcs = bc_set.boundary_conditions
        mesh = field.mesh
        tri_ids = mesh.get_triangles(port.tags)
        logger.trace(f'.boundary face has {len(tri_ids)} triangles.')

        origin = tuple([c-n for c,n in zip(port.cs.origin, port.cs.gzhat)])
        logger.trace(f'.boundary origin {origin}')

        boundary_surface = mesh.boundary_surface(port.tags, origin)
        nedlegfield = NedelecLegrange2(boundary_surface, port.cs)

        ermesh = er[:,:,tri_ids]
        urmesh = ur[:,:,tri_ids]
        sigmesh = sig[tri_ids]
        ermesh = ermesh - 1j * sigmesh/(k0*C0*EPS0)

        logger.trace(f'.assembling matrices for {nedlegfield} at k0={k0:.2f}')
        E, B = generelized_eigenvalue_matrix(nedlegfield, ermesh, urmesh, port.cs._basis, k0)

        # TODO: Simplified to all "conductors" loosely defined. Must change to implementing line robin boundary conditions.
        pecs: list[BoundaryCondition] = bc_set.get_conductors()#[bc for bc in bcs if isinstance(bc,PEC)]
        if len(pecs) > 0:
            logger.debug(f'.total of equiv. {len(pecs)} PEC BCs implemented')

        pec_ids = []

        # Process all concutors. Everything above the conductivity limit is considered pec.
        for it in range(boundary_surface.n_tris):
            if sigmesh[it] > self.conductivity_limit:
                pec_ids.extend(list(nedlegfield.tri_to_field[:,it]))

        # Process all PEC Boundary Conditions
        for pec in pecs:
            logger.trace(f'.implementing {pec}')
            face_tags = pec.tags
            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())
            for ii in edge_ids:
                i2 = nedlegfield.mesh.from_source_edge(ii)
                if i2 is None:
                    continue
                eids = nedlegfield.edge_to_field[:, i2]
                pec_ids.extend(list(eids))

        # Process all port boundary Conditions
        pec_ids_set: set[int] = set(pec_ids)

        logger.trace(f'.total of {len(pec_ids_set)} pec DoF to remove.')
        solve_ids = [i for i in range(nedlegfield.n_field) if i not in pec_ids_set]

        return E, B, np.array(solve_ids), nedlegfield

    def assemble_freq_matrix(self, field: Nedelec2, 
                        materials: list[Material],
                        bcs: list[BoundaryCondition],
                        frequency: float,
                        cache_matrices: bool = False) -> SimJob:
        """Assembles the frequency domain FEM matrix

        Args:
            field (Nedelec2): The Nedelec2 object of the problems
            er (np.ndarray): The relative dielectric permitivity tensor of shape (3,3,N)
            ur (np.ndarray): The relative magnetic permeability tensor of shape (3,3,N)
            sig (np.ndarray): The conductivity array of shape (N,)
            bcs (list[BoundaryCondition]): The boundary conditions
            frequency (float): The simulation frequency
            cache_matrices (bool, optional): Whether to use and cache matrices. Defaults to False.

        Returns:
            SimJob: The resultant SimJob object
        """

        from .curlcurl import tet_mass_stiffness_matrices
        from .robinbc import assemble_robin_bc, assemble_robin_bc_excited

        # PREDEFINE CONSTANTS
        W0 = 2*np.pi*frequency
        K0 = W0/C0
        
        is_frequency_dependent = False
        mesh = field.mesh

        for mat in materials:
            if mat.frequency_dependent:
                is_frequency_dependent = True
                break

        er = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        tand = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        cond = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        ur = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        
        for mat in materials:
            er = mat.er(frequency, er)
            ur = mat.ur(frequency, ur)
            tand = mat.tand(frequency, tand)
            cond = mat.cond(frequency, cond)
        
        er = er*(1-1j*tand) - 1j*cond/(W0*EPS0)
        
        is_frequency_dependent = is_frequency_dependent or np.any((cond > 0) & (cond < self.conductivity_limit)) # type: ignore

        if cache_matrices and not is_frequency_dependent and self.cached_matrices is not None:
            # IF CACHED AND AVAILABLE PULL E AND B FROM CACHE
            logger.debug('Using cached matricies.')
            E, B = self.cached_matrices
        else:
            # OTHERWISE, COMPUTE
            logger.debug('Assembling matrices')
            E, B = tet_mass_stiffness_matrices(field, er, ur)
            self.cached_matrices = (E, B)

        # COMBINE THE MASS AND STIFFNESS MATRIX
        K: csr_matrix = (E - B*(K0**2)).tocsr()

        NF = E.shape[0]

        # ISOLATE BOUNDARY CONDITIONS TO ASSEMBLE
        pec_bcs: list[PEC] = [bc for bc in bcs if isinstance(bc,PEC)]
        robin_bcs: list[RobinBC] = [bc for bc in bcs if isinstance(bc,RobinBC)]
        port_bcs: list[PortBC] = [bc for bc in bcs if isinstance(bc, PortBC)]
        periodic_bcs: list[Periodic] = [bc for bc in bcs if isinstance(bc, Periodic)]

        # PREDEFINE THE FORCING VECTOR CONTAINER
        b = np.zeros((E.shape[0],)).astype(np.complex128)
        port_vectors = {port.port_number: np.zeros((E.shape[0],)).astype(np.complex128) for port in port_bcs}
        

        ############################################################
        #                      PEC BOUNDARY CONDITIONS             #
        ############################################################

        logger.debug('Implementing PEC Boundary Conditions.')
        pec_ids: list[int] = []
        
        # Conductivity above al imit, consider it all PEC
        ipec = 0
        for itet in range(field.n_tets):
            if cond[0,0,itet] > self.conductivity_limit:
                ipec+=1
                pec_ids.extend(field.tet_to_field[:,itet])
        if ipec>0:
            logger.trace(f'Extended PEC with {ipec} tets with a conductivity > {self.conductivity_limit}.')

        for pec in pec_bcs:
            logger.trace(f'Implementing: {pec}')
            if len(pec.tags)==0:
                continue
            face_tags = pec.tags
            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

            for ii in edge_ids:
                eids = field.edge_to_field[:, ii]
                pec_ids.extend(list(eids))

            for ii in tri_ids:
                tids = field.tri_to_field[:, ii]
                pec_ids.extend(list(tids))


        ############################################################
        #                     ROBIN BOUNDARY CONDITIONS            #
        ############################################################

        if len(robin_bcs) > 0:
            logger.debug('Implementing Robin Boundary Conditions.')
        
            gauss_points = gaus_quad_tri(4)
            Bempty = field.empty_tri_matrix()
            for bc in robin_bcs:
                logger.trace(f'.Implementing {bc}')
                for tag in bc.tags:
                    face_tags = [tag,]

                    tri_ids = mesh.get_triangles(face_tags)
                    nodes = mesh.get_nodes(face_tags)
                    edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

                    gamma = bc.get_gamma(K0)
                    logger.trace(f'..robin bc γ={gamma:.3f}')

                    def Ufunc(x,y,z): 
                        return bc.get_Uinc(x,y,z,K0)
                    
                    if bc._include_force:
                        Bempty, b_p = assemble_robin_bc_excited(field, Bempty, tri_ids, Ufunc, gamma, gauss_points) # type: ignore
                        port_vectors[bc.port_number] += b_p # type: ignore
                        logger.trace(f'..included force vector term with norm {np.linalg.norm(b_p):.3f}')
                    else:
                        Bempty = assemble_robin_bc(field, Bempty, tri_ids, gamma) # type: ignore
            B_p = field.generate_csr(Bempty)
            K = K + B_p
        
        if len(periodic_bcs) > 0:
            logger.debug('Implementing Periodic Boundary Conditions.')


        ############################################################
        #                   PERIODIC BOUNDARY CONDITIONS          #
        ############################################################

        Pmats = []
        remove: set[int] = set()
        has_periodic = False

        for pbc in periodic_bcs:
            logger.trace(f'.Implementing {pbc}')
            has_periodic = True
            tri_ids_1 = mesh.get_triangles(pbc.face1.tags)
            edge_ids_1 = mesh.get_edges(pbc.face1.tags)
            tri_ids_2 = mesh.get_triangles(pbc.face2.tags)
            edge_ids_2 = mesh.get_edges(pbc.face2.tags)
            dv = np.array(pbc.dv)
            logger.trace(f'..displacement vector {dv}')
            linked_tris = pair_coordinates(mesh.tri_centers, tri_ids_1, tri_ids_2, dv, 1e-9)
            linked_edges = pair_coordinates(mesh.edge_centers, edge_ids_1, edge_ids_2, dv, 1e-9)
            dv = np.array(pbc.dv)
            phi = pbc.phi(K0)
            logger.trace(f'..ϕ={phi} rad/m')
            Pmat, rows = gen_periodic_matrix(tri_ids_1,
                                       edge_ids_1,
                                       field.tri_to_field,
                                       field.edge_to_field,
                                       linked_tris,
                                       linked_edges,
                                       field.n_field,
                                       phi)
            remove.update(rows)
            Pmats.append(Pmat)

        if Pmats:
            logger.trace(f'.periodic bc removes {len(remove)} boundary DoF')
            Pmat = Pmats[0]
            for P2 in Pmats[1:]:
                Pmat = Pmat @ P2
            remove_array = np.sort(np.unique(list(remove)))
            all_indices = np.arange(NF)
            keep_indices = np.setdiff1d(all_indices, remove_array)
            Pmat = Pmat[:,keep_indices]
        else:
            Pmat = None
        

        ############################################################
        #                             FINALIZE                     #
        ############################################################

        pec_ids_set = set(pec_ids)
        solve_ids = np.array([i for i in range(E.shape[0]) if i not in pec_ids_set])
        
        if has_periodic:
            mask = np.zeros((NF,))
            mask[solve_ids] = 1
            mask = mask[keep_indices]
            solve_ids = np.argwhere(mask==1).flatten()

        logger.debug(f'Number of tets: {mesh.n_tets:,}')
        logger.debug(f'Number of DoF: {K.shape[0]:,}')
        logger.debug(f'Number of non-zero: {K.nnz:,}')
        simjob = SimJob(K, b, K0*299792458/(2*np.pi), True)
        
        simjob.port_vectors = port_vectors
        simjob.solve_ids = solve_ids

        if has_periodic:
            simjob.P = Pmat
            simjob.Pd = Pmat.getH()
            simjob.has_periodic = has_periodic

        return simjob, (er, ur, cond)
    
    def assemble_eig_matrix(self, field: Nedelec2, 
                        materials: list[Material],
                        bcs: list[BoundaryCondition],
                        frequency: float) -> SimJob:
        """Assembles the eigenmode analysis matrix

        The assembly process is frequency dependent because the frequency-dependent properties
        need a guess before solving. There is currently no adjustment after an eigenmode is found.
        The frequency-dependent properties are simply calculated once for the given frequency

        Args:
            field (Nedelec2): The Nedelec2 field
            er (np.ndarray): The relative permittivity tensor in shape (3,3,N)
            ur (np.ndarray): The relative permeability tensor in shape (3,3,N)
            sig (np.ndarray): The conductivity scalar in array (N,)
            bcs (list[BoundaryCondition]): The list of boundary conditions
            frequency (float): The compilation frequency (for material properties only)

        Returns:
            SimJob: The resultant simulation job
        """
        from .curlcurl import tet_mass_stiffness_matrices
        from .robinbc import assemble_robin_bc

        mesh = field.mesh
        w0 = 2*np.pi*frequency
        k0 = w0/C0

        er = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        tand = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        cond = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        ur = np.zeros((3,3,field.mesh.n_tets), dtype=np.complex128)
        
        for mat in materials:
            er = mat.er(frequency, er)
            ur = mat.ur(frequency, ur)
            tand = mat.tand(frequency, tand)
            cond = mat.cond(frequency, cond)
        
        er = er*(1-1j*tand) - 1j*cond/(w0*EPS0)
        
        logger.debug('Assembling matrices')
        
        E, B = tet_mass_stiffness_matrices(field, er, ur)
        self.cached_matrices = (E, B)

        NF = E.shape[0]

        pecs: list[PEC] = [bc for bc in bcs if isinstance(bc,PEC)]
        robin_bcs: list[RectangularWaveguide] = [bc for bc in bcs if isinstance(bc,RobinBC)] # type: ignore
        periodic: list[Periodic] = [bc for bc in bcs if isinstance(bc, Periodic)]

        # Process all PEC Boundary Conditions
        pec_ids: list = []
        
        logger.debug('Implementing PEC Boundary Conditions.')
        
        # Conductivity above a limit, consider it all PEC
        for itet in range(field.n_tets):
            if cond[0,0,itet] > self.conductivity_limit:
                pec_ids.extend(field.tet_to_field[:,itet])
        
        # PEC Boundary conditions
        for pec in pecs:
            if len(pec.tags)==0:
                continue
            face_tags = pec.tags
            tri_ids = mesh.get_triangles(face_tags)
            edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

            for ii in edge_ids:
                eids = field.edge_to_field[:, ii]
                pec_ids.extend(list(eids))

            for ii in tri_ids:
                tids = field.tri_to_field[:, ii]
                pec_ids.extend(list(tids))

        # Robin BCs
        if len(robin_bcs) > 0:
            logger.debug('Implementing Robin Boundary Conditions.')
            
            Bempty = field.empty_tri_matrix()
            for bc in robin_bcs:

                for tag in bc.tags:
                    face_tags = [tag,]

                    tri_ids = mesh.get_triangles(face_tags)
                    nodes = mesh.get_nodes(face_tags)
                    edge_ids = list(mesh.tri_to_edge[:,tri_ids].flatten())

                    gamma = bc.get_gamma(k0)
                    
                    ibasis = bc.get_inv_basis()
                    if ibasis is None:
                        basis = plane_basis_from_points(mesh.nodes[:,nodes]) + 1e-16
                        ibasis = np.linalg.pinv(basis)
                    
                    Bempty = assemble_robin_bc(field, Bempty, tri_ids, gamma) # type: ignore
            B_p = field.generate_csr(Bempty)
            B = B + B_p
        
        if len(periodic) > 0:
            logger.debug('Implementing Periodic Boundary Conditions.')

        # Periodic BCs
        Pmats = []
        remove = set()
        has_periodic = False

        for bcp in periodic:
            has_periodic = True
            tri_ids_1 = mesh.get_triangles(bcp.face1.tags)
            edge_ids_1 = mesh.get_edges(bcp.face1.tags)
            tri_ids_2 = mesh.get_triangles(bcp.face2.tags)
            edge_ids_2 = mesh.get_edges(bcp.face2.tags)
            dv = np.array(bcp.dv)
            linked_tris = pair_coordinates(mesh.tri_centers, tri_ids_1, tri_ids_2, dv, 1e-9)
            linked_edges = pair_coordinates(mesh.edge_centers, edge_ids_1, edge_ids_2, dv, 1e-9)
            dv = np.array(bcp.dv)
            phi = bcp.phi(k0)
            
            Pmat, rows = gen_periodic_matrix(tri_ids_1,
                                       edge_ids_1,
                                       field.tri_to_field,
                                       field.edge_to_field,
                                       linked_tris,
                                       linked_edges,
                                       field.n_field,
                                       phi)
            remove.update(rows)
            Pmats.append(Pmat)
        
        if Pmats:
            Pmat = Pmats[0]
            for P2 in Pmats[1:]:
                Pmat = Pmat @ P2
            Pmat = Pmat.tocsr()
            remove_array = np.sort(np.array(list(remove)))
            all_indices = np.arange(NF)
            keep_indices = np.setdiff1d(all_indices, remove_array)
            Pmat = Pmat[:,keep_indices]
        else:
            Pmat = None
        
        pec_ids_set = set(pec_ids)
        solve_ids = np.array([i for i in range(E.shape[0]) if i not in pec_ids_set])
        
        if has_periodic:
            mask = np.zeros((NF,))
            mask[solve_ids] = 1
            mask = mask[keep_indices]
            solve_ids = np.argwhere(mask==1).flatten()

        logger.debug(f'Number of tets: {mesh.n_tets}')
        logger.debug(f'Number of DoF: {E.shape[0]}')
        simjob = SimJob(E, None, frequency, False, B=B)
        
        simjob.solve_ids = solve_ids

        if has_periodic:
            simjob.P = Pmat
            simjob.Pd = Pmat.getH()
            simjob.has_periodic = has_periodic

        return simjob, (er, ur, cond)


import emerge as em

""" DIELECTRIC RESONATOR FILTER DEMO

This demo constructs and analyzes a dielectric resonator filter using EMerge.
A dielectric cylinder (resonator) sits on a supporting block within a metal
enclosure. We solve for its resonant modes via eigenmode analysis and visualize
electric and magnetic field distributions for each mode. """

# --- Unit definitions -----------------------------------------------------
mm = 0.001                       # meter per millimeter
inch = 25.4 * mm                 # meter per inch

# --- Geometry dimensions --------------------------------------------------
S = 2.03 * inch                  # enclosure height
W = 2.0 * inch                   # enclosure width/length (square base)

# Supporting block (substrate) dimensions
Dsup = 0.56 * inch               # support cylinder diameter
Lsup = 0.8 * inch                # support cylinder height
# Resonator cylinder dimensions
Dres = 1.176 * inch              # resonator cylinder diameter
Lres = 0.481 * inch              # resonator cylinder height

# --- Material definitions ------------------------------------------------
# High-er support material (e.g., alumina)
mat_support = em.lib.Material(er=10, color="#ffffff")
# Dielectric resonator material (e.g., ceramic)
mat_resonator = em.lib.Material(er=34, color="#ededed")

# Number of resonant modes to extract
Nmodes = 5

# --- Create simulation ---------------------------------------------------
model = em.Simulation('DielectricResonatorFilter')
model.check_version("0.6.8") # Checks version compatibility.

# --- Build geometry ------------------------------------------------------
# Metal enclosure box (PEC by default)
box = em.geo.Box(
    W, W, S,
    position=(-W/2, -W/2, 0)
)
# Support cylinder centered on enclosure floor
support = em.geo.Cylinder(
    radius=Dsup/2,
    height=Lsup,
    cs=em.GCS,
    Nsections=20
).set_material(mat_support).prio_up()

# DDR cylinder placed atop support
resonator = em.geo.Cylinder(
    radius=Dres/2,
    height=Lres,
    cs=em.GCS.displace(0, 0, Lsup),
    Nsections=32
).set_material(mat_resonator).prio_up()

# Assemble geometry into model
model.commit_geometry()

# --- Solver settings -----------------------------------------------------
# Only eigenmode solver needed; set center frequency estimate
model.mw.set_frequency(3e9)

# --- Mesh generation -----------------------------------------------------
model.generate_mesh()

# --- Eigenmode analysis --------------------------------------------------
# Solve for the first Nmodes resonant frequencies around 2 GHz
data = model.mw.eigenmode(
    2e9,
    nmodes=Nmodes
)

# --- Visualization of modes ----------------------------------------------
for mode_index in range(Nmodes):
    # Extract field grid for this mode (sample spacing ~0.2 in)
    field = data.field[mode_index].grid(0.2 * inch)
    # Show enclosure, support, and resonator transparently
    model.display.add_object(box, opacity=0.1)
    model.display.add_object(support, opacity=0.3)
    model.display.add_object(resonator, opacity=0.3)
    # Plot E-field vectors in red and H-field vectors in blue
    Evec = field.vector('E', 'real')
    Hvec = field.vector('H', 'real')
    model.display.add_quiver(*Evec, color='red')
    model.display.add_quiver(*Hvec, color='blue')
    # Annotate resonant frequency and field labels
    freq_ghz = data.field[mode_index].freq.real / 1e9
    model.display.add_title(f'Mode {mode_index+1}: {freq_ghz:.3f} GHz')
    model.display.add_text('E-field', color='red', abs_position=(0, 0.95))
    model.display.add_text('H-field', color='blue', abs_position=(0, 0.9))
    model.display.add_surf(*data.field[mode_index].cutplane(2*mm, y=0).scalar('normS','real'))
    # Render each mode one at a time
    model.display.show()



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from typing import TypeVar, Generic, Any
from loguru import logger
from .physics.microwave.microwave_data import MWData
from .simulation_data import DataContainer, DataEntry

class SimulationDataset:
    """This simple class contains the different kinds of data sets in the Simulation Model. It includes

    Attributes:
      self.mw: MWData              - The Microwave physics data
      self.globals: dict[str, Any] - Any globally defined data of choice in the Simulation
      self.sim: DataContainer      - Generic simulation data associated with different instantiation of your at parameter level.
    """
    def __init__(self):
        self.mw: MWData = MWData()
        self.globals: dict[str, Any] = dict()
        self.sim: DataContainer = DataContainer()
        


import emerge as em
from emerge.plot import plot_sp, plot_ff
import numpy as np

""" STANDARD GAIN HORN ANTENNA

This demo sets up and simulates a rectangular horn antenna in an
absorbing domain with PML layers. We compute return loss (S11) over a
90–110 GHz band and plot the far-field radiation pattern. 

The dimensions come from this paper:
https://pure.tue.nl/ws/portalfiles/portal/332971061/Uncertainties_in_the_Estimation_of_the_Gain_of_a_Standard_Gain_Horn_in_the_Frequency_Range_of_90_GHz_to_140_GHz.pdf
"""

# --- Units ---------------------------------------------------------------
mm = 0.001               # meters per millimeter

# --- Horn and feed dimensions -------------------------------------------
wga = 2.01 * mm           # waveguide width
wgb = 1.01 * mm          # waveguide height
WH = 10 * mm            # Aperture width
HH = 7 * mm             # Aperture height
Lhorn = 21 * mm        # Horn length

# --- Feed and simulation setup ------------------------------------------
Lfeed = 2 * mm           # length of feed waveguide
th = 1 * mm            # PML thickness
dx = 2 * mm              # distance from horn exit to PML start

# Create simulation object
m = em.Simulation('HornAntenna')
m.check_version("0.6.8") # Checks version compatibility.
# --- Coordinate system for horn geometry -------------------------------
hornCS = em.CS(em.YAX, em.ZAX, em.XAX)

# Feed waveguide as rectangular box (metal)
feed = em.geo.Box(
    Lfeed,   # length along X
    wga/2,     # half-width along Y (centered)
    wgb/2,     # half-height along Z
    position=(-Lfeed, 0, 0)
)

# --- Horn geometry ------------------------------------------------------
# Inner horn taper from (D,E) at throat to (B,C) at mouth over length F
horn_in = em.geo.Horn(
    (wga, wgb), (WH, HH), Lhorn, hornCS,
)
# Outer horn (including metal thickness) helps define PML subtraction
horn_out = em.geo.Horn(
    (wga+2*th, wgb+2*th), (WH+2*th, HH+2*th), Lhorn, hornCS
)

# --- Bounding objects and PML -------------------------------------------
# Define large intersection box to trim horn geometry
ibox = em.geo.Box(30*mm, 30*mm, 30*mm)
horn_in = em.geo.intersect(horn_in, ibox, remove_tool=False)
horn_out = em.geo.intersect(horn_out, ibox)

# Create airbox with PML layers on +X, +Y, +Z faces
rat = 1.6  # PML extension ratio
air, *pmls = em.geo.pmlbox(
    4*mm,               # air padding before PML
    rat*WH/2,           # half-height in Y
    rat*HH/2,           # half-width in Z
    (Lhorn - dx, 0, 0), # PML origin offset along X
    thickness=4*mm,
    N_mesh_layers=4,
    sides='TRA'         # [T]op, [R]ight, B[A]ck
)
# Subtract horn volume from airbox so PML does not cover metal
air2 = em.geo.subtract(air, horn_out)

# --- Solver parameters --------------------------------------------------
m.mw.set_frequency_range(90e9, 110e9, 11)  # 90–110 GHz sweep
m.mw.set_resolution(0.25)                  # mesh resolution fraction

# --- Assemble geometry and mesh -----------------------------------------
m.generate_mesh()

# --- Boundary conditions ------------------------------------------------
p1 = m.mw.bc.ModalPort(feed.face('left'), 1)     # excite TE10 in feed
PMC = m.mw.bc.PMC(m.select.face.inplane(0, 0, 0, 0, 1, 0))  # perfect magnetic on symmetry
radiation_boundary = air2.faces('back','top','right', tool=air)  # open faces
abc = m.mw.bc.AbsorbingBoundary(m.select.face.inplane(Lhorn-dx,0,0,1,0,0))

# View mesh and BC selections
m.view(selections=[p1.selection, PMC.selection, radiation_boundary])

# --- Run frequency-domain solver ----------------------------------------
data = m.mw.run_sweep()

# --- Plot return loss ---------------------------------------------------
scal = data.scalar.grid
plot_sp(scal.freq, scal.S(1,1), labels=['S11'])  # S11 vs frequency

# --- Far-field radiation pattern ----------------------------------------
# Compute E and H on 2D cut for phi=0 plane over -90° to 90°
ff_data = data.field[0].farfield_2d(
    (1, 0, 0), (0, 1, 0), radiation_boundary,
    (-90, 90), syms=['Ez','Hy']
)

plot_ff(ff_data.ang * 180/np.pi, ff_data.normE/em.lib.EISO, dB=True, ylabel='Gain [dBi]')
# Normalize to free-space impedance and convert to dB

m.display.add_object(horn_in, opacity=0.1)
m.display.add_object(air2, opacity=0.1)
m.display.add_object(feed, opacity=0.1)
m.display.add_surf(*data.field[0].farfield_3d(radiation_boundary, syms=['Ez','Hy'])\
                .surfplot('normE', 'abs', True, True, -30, 5*mm, (Lhorn,0,0)), cmap='viridis', symmetrize=False)
m.display.add_surf(*data.field[0].cutplane(0.25*mm, z=0).scalar('Ez','real'))
m.display.show()




import emerge as em
from emerge.plot import plot_sp

"""PCB Vias

This demonstration shows how to add vias with the PCB router. Make sure to go through the other
PCB related demos (demo1 and demo3) to get more information on the PCBLayouter.

Notice that the results of this simulation are not supposed to be good. Its more about the geometry than the S-parameters.

"""

mm = 0.001    # Define a millimeter
th = 1.0      # mm

model = em.Simulation('Stripline_test')
model.check_version("0.6.8") # Checks version compatibility.

# As usual we start by creating our layouter
ly = em.geo.PCB(th, mm, em.GCS, em.lib.DIEL_RO4350B)

# Here we define a simple stripline path that makes a knick turn and a via jump to a new layer.
# None of the transmission lines are conciously matched in any way, this is just about the routing

# The .via(...) method allows one to add a via geometry to the PCBLayouter that can be extracted later
# A user may decide whether to proceed or not beyond the via. More information can be found in the 
# docstring of the method. The vias can be created later.
ly.new(0,0,1,(1,0), -th/2).store('p1').straight(10).turn(90).straight(10).turn(-90)\
    .straight(2).via(0, 0.2, True).straight(8).via(-th/2, 0.2).straight(2)\
        .turn(-90).straight(10).turn(90).straight(10).store('p2')

# As usual we compile the traces as a merger of polygons
trace = ly.compile_paths(True)

# Now that we have via's defined, we can do the same with vias. I set Merge to True so that I get back
# One GeoObject.
vias = ly.generate_vias(True)

# Here I use lumped ports instead of wave ports. I use the references made earlier to generate the port.
# By default, all lumped port sheets will be shorted to z=-thickness. You can change this as an optional
# argument.
lp1 = ly.lumped_port(ly.load('p1'))
lp2 = ly.lumped_port(ly.load('p2'))

# Because lumped ports don't stop at the edge of our domain, we make sure to add some margins everywhere.
ly.determine_bounds(5,5,5,5)

# Finally we can generate the PCB volumes. Because the trace start halfway through the PCB we turn
# on the split-z function which cuts the PCB in multiple layers. This improves meshing around the striplines.
diel = ly.generate_pcb(True, merge=True)

# We also define the air-box with 3mm thickness
air = ly.generate_air(3.0)

# Finish modelling by calling commit_geometry
model.commit_geometry()

model.view()

model.mw.set_frequency_range(1e9, 6e9, 11)
model.mesher.set_boundary_size(trace, 1*mm)

model.generate_mesh()

# We display the geometry with extra attention to the vias. With the vias.boundary() method we can
# specifically show the outside faces of the via.
model.view(selections=[vias.boundary()])

# We setup the lumped port boundary conditions. Because of an added functionality in the PCBLayouter 
# class, you don't have to specify the width, height and direction of the lumped port, this information
# is contained in the lumped port sheet. You can see this information as its stored in the lp1._aux_data
# dictionary.

p1 = model.mw.bc.LumpedPort(lp1, 1)
p2 = model.mw.bc.LumpedPort(lp2, 2)
pec = model.mw.bc.PEC(trace)
pecvia = model.mw.bc.PEC(vias.boundary())

# Finally we run the simulation!
data = model.mw.run_sweep(True, 4, frequency_groups=8)

freq = data.scalar.grid.freq
S11 = data.scalar.grid.S(1,1)
S21 = data.scalar.grid.S(2,1)

plot_sp(freq, [S11, S21], labels=['S11','S21'])

model.display.add_object(diel, opacity=0.2)
model.display.add_object(trace)
model.display.add_object(vias)

# You can use the cutplane method of the BaseDataset class
# This is equivalent to the interpolate method except it automatically generates
# the point cloud based on a plane x,y or z coordinate.
model.display.add_quiver(*data.field[3].cutplane(ds=0.001, z=-0.00025).vector('E'))
model.display.add_surf(*data.field[3].cutplane(ds=0.001, z=-0.00075).scalar('Ez','real'))
model.display.show()


import emerge as em
"""This demo is still in progress. 

For now it just shows you how to work with the revolve system.
"""
model = em.Simulation('Revolve test')
model.check_version("0.6.8") # Checks version compatibility.

mm = 0.001
rad_feed = 30*mm
rad_out = 70*mm
len_horn = 100*mm
len_feed = 40*mm
th = 5*mm

poly = em.geo.XYPolygon([rad_feed, rad_feed, rad_out, 0, 0], [-len_feed, 0, len_horn, len_horn, -len_feed])
vol_in = poly.revolve(em.ZXPLANE.cs(), (0,0,0), (1,0,0))
poly = em.geo.XYPolygon([rad_feed+th, rad_feed+th, rad_out+th, 0, 0], [-len_feed, 0, len_horn, len_horn, -len_feed])
vol_out = poly.revolve(em.ZXPLANE.cs(), (0,0,0), (1,0,0))

ratio = 2.5
airbox = em.geo.pmlbox(40*mm, ratio*rad_out, ratio*rad_out, (len_horn-5*mm, -ratio*rad_out/2, -ratio*rad_out/2), thickness=30*mm, 
                       top=True, bottom=True, right=True, front=True, back=True)

model.view()



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
from enum import Enum
from loguru import logger
from .selection import Selection, FaceSelection
import numpy as np
from .geometry import GeoObject
from typing import TypeVar, Type

T = TypeVar('T')

class BCDimension(Enum):
    ANY = -1
    NODE = 0
    EDGE = 1
    FACE = 2
    DOMAIN = 3

def _unique(input: list[int]) -> list:
    """ Returns a sorted list of all unique integers/floats in a list."""
    return sorted(list(set(input)))

############################################################
#               BASE BOUNDARY CONDITION CLASS              #
############################################################

class BoundaryCondition:
    """A generalized class for all boundary condition objects.
    """
    
    def __init__(self, assignment: GeoObject | Selection):

        self.dimension: BCDimension = BCDimension.ANY
        self.indices: list[int] = []
        self.face_indices: list[int] = []
        self.edge_indices: list[int] = []
        
        
        if isinstance(assignment, GeoObject):
            assignment = assignment.select
        
        self.selection: Selection = assignment
        self.tags: list[int] = self.selection.tags
    
    @property
    def dim(self) -> int:
        ''' The dimension of the boundary condition as integer (0,1,2,3).'''
        return self.dimension.value
    
    def __repr__(self) -> str:
        return f'{type(self).__name__}{self.tags}'

    def __str__(self) -> str:
        return self.__repr__()
    
    def check_dimension(self, tags: list[tuple[int,int]]) -> None:
        # check if all tags have the same timension (dim, tag)
        if not isinstance(tags, list):
            raise TypeError(f'Argument tags must be of type list, instead its {type(tags)}')
        if len(tags) == 0:
            return
        if not all(isinstance(x, tuple) and len(x) == 2 for x in tags):
            raise TypeError(f'Argument tags must be of type list of tuples, instead its {type(tags)}')
        if not all(isinstance(x[0], int) and isinstance(x[1], int) for x in tags):
            raise TypeError(f'Argument tags must be of type list of tuples of ints, instead its {type(tags)}')
        if not all(x[0] == tags[0][0] for x in tags):
            raise ValueError(f'All tags must have the same dimension, instead its {tags}')
        dimension = tags[0][0]
        if self.dimension is BCDimension.ANY:
            logger.info(f'Assigning dimension {BCDimension(dimension)} to {self}')
            self.dimension = BCDimension(dimension)
        elif self.dimension != BCDimension(dimension):
            raise ValueError(f'Current boundary condition has dimension {self.dimension}, but tags have dimension {BCDimension(dimension)}')
        
    def add_tags(self, dimtags: list[tuple[int,int]]) -> None:
        """Adds the given taggs to this boundary condition.

        Args:
            tags (list[tuple[int,int]]): The tags to include
        """
        self.check_dimension(dimtags)
        tags = [x[1] for x in dimtags]
        self.tags = _unique(self.tags + tags)
    
    def remove_tags(self, tags: list[int]) -> list[int]:
        """Removes the tags provided by tags from this boundary condition.

        Return sonly the tags that are actually excluded from this face.

        Args:
            tags (list[int]): The tags to exclude.

        Returns:
            list[int]: A list of actually excluded tags.
        """
        excluded_edges = [x for x in self.tags if x in tags]
        self.tags = [x for x in self.tags if x not in tags]
        return excluded_edges
    
    def exclude_bc(self, other: BoundaryCondition) -> list[int]:
        """Excludes all faces for a provided boundary condition object from this boundary condition assignment.

        Args:
            other (BoundaryCondition): The boundary condition of which the faces should be excluded

        Returns:
            list[int]: A list of excluded face tags.
        """
        return self.remove_tags(other.tags)

class BoundaryConditionSet:

    def __init__(self):

        self.boundary_conditions: list[BoundaryCondition] = []
        self._initialized: bool = False
    
    def _construct_bc(self, constructor: type) -> type:
        def constr(*args, **kwargs):
            obj = constructor(*args, **kwargs)
            self.assign(obj)
            return obj
        return constr # type: ignore
    
    def assigned(self, dim: int = 2) -> list[int]:
        """Returns all boundary tags that have a boundary condition assigned to them

        Args:
            dim (int, optional): The dimension. Defaults to 2.

        Returns:
            list[int]: The list of tags
        """
        tags = []
        for bc in self.boundary_conditions:
            if bc.dim != dim:
                continue
            tags.extend(bc.tags)
        return tags

    def count(self, bctype: type) -> int:
        """Returns the number of a certain boundary condition type

        Args:
            bctype (type): The boundary condtion type

        Returns:
            int: The number of occurances
        """
        return len(self.oftype(bctype))
    
    def oftype(self, bctype: Type[T]) -> list[T]:
        """Returns a list of all boundary conditions of a certain type.

        Args:
            bctype (type): The boundar condition type

        Returns:
            list[BoundaryCondition]: The list of boundary conditions
        """
        return [item for item in self.boundary_conditions if isinstance(item, bctype)]

    def reset(self) -> None:
        """Resets the boundary conditions that are defined
        """
        self.boundary_conditions = []

    def assign(self, 
               bc: BoundaryCondition) -> None:
        """Assign a boundary-condition object to a domain or list of domains.
        This method must be called to submit any boundary condition object you made to the physics.

        Args:
            bcs *(BoundaryCondition): A list of boundary condition objects.
        """
        if bc in self.boundary_conditions:
            return
        self._initialized = True
        wordmap = {
            0: 'point',
            1: 'edge',
            2: 'face',
            3: 'domain'
        }

        bc.add_tags(bc.selection.dimtags)

        logger.info('Excluding other possible boundary conditions')

        for existing_bc in self.boundary_conditions:
            excluded = existing_bc.exclude_bc(bc)
            if excluded:
                logger.debug(f'Removed the {excluded} tags from {wordmap[bc.dim]} BC {existing_bc}')
        self.boundary_conditions.append(bc)

class Periodic(BoundaryCondition):

    def __init__(self, 
                 selection1: Selection,
                 selection2: Selection,
                 dv: tuple[float,float,float],
                 ):
        self.face1: BoundaryCondition = BoundaryCondition(selection1)
        self.face2: BoundaryCondition = BoundaryCondition(selection2)
        super().__init__(FaceSelection(selection1.tags + selection2.tags))
        self.dv: tuple[float,float,float] = dv
        self.ux: float = 0
        self.uy: float = 0
        self.uz: float = 0

    def phi(self, k0) -> complex:
        dx, dy, dz = self.dv
        return np.exp(-1j*k0*(self.ux*dx+self.uy*dy+self.uz*dz))
    


import emerge as em
from emerge.plot import plot_sp

""" DEMO 12: Mode Alignment

In some applications we wish to simulate transmission lines with multiple modes. In this
example we look at a circular waveguide with 4 ridges. The modal analysis performed to find
the modes are not guaranteed however to return each mode in order, nor in the right polarization
direction (+Z vs -Z). In EMerge we can alignm modes using the .align_mode() method.

"""

# First we define our simulation
m = em.Simulation('aligntest')
m.check_version("0.6.8") # Checks version compatibility.

# We create a cyllindrical waveguide in the Y-axis.
cyl = em.geo.Cylinder(0.012, 0.05, em.CS(em.XAX, em.ZAX, em.YAX))

# We remove 4 ridges from this with a width of 2mm. 
wr= 0.002
gap = 0.015
ridge1 = em.geo.Box(wr, 0.05, 0.02, position=(-wr/2, 0, gap/2))
ridge2 = em.geo.Box(wr, 0.05, 0.02, position=(-wr/2, 0, gap/2))
ridge3 = em.geo.Box(wr, 0.05, 0.02, position=(-wr/2, 0, gap/2))
ridge4 = em.geo.Box(wr, 0.05, 0.02, position=(-wr/2, 0, gap/2))

ridge2 = em.geo.rotate(ridge2, (0,0,0), (0,1,0), 90)
ridge3 = em.geo.rotate(ridge3, (0,0,0), (0,1,0), 180)
ridge4 = em.geo.rotate(ridge4, (0,0,0), (0,1,0), 270)

wg = em.geo.subtract(cyl, ridge1)
wg = em.geo.subtract(wg, ridge2)
wg = em.geo.subtract(wg, ridge3)
wg = em.geo.subtract(wg, ridge4)


# We define a resolution and mesh our geometry
m.mw.set_resolution(0.2)
m.mw.set_frequency_range(8e9,10e9, 7)

m.generate_mesh()

m.view()

# Next we define our modal ports.
p1 = m.mw.bc.ModalPort(wg.face('front', tool=cyl), 1)
p2 = m.mw.bc.ModalPort(wg.face('back', tool=cyl), 2)

# To align our modes we use the .align_modes() method. This takes a series of arguments
# in the form of an Axis object, a tuple like (1,0,0) for +X or a numpy array (np.array([0,1,0])).ArithmeticError
# After solving for the modes, each mode will be sorted based on which mode has the largest inner product with these axis objects.
# The axes will also be used to align the phase (0 degrees vs 180 degrees).
# Because the waveguide supports two modes in the X and Z polarization direction we will use these axes to align them.ArithmeticError

p1.align_modes(em.ZAX, em.XAX)
p2.align_modes(em.ZAX, em.XAX)

# We will now iterate through our modes and solve our system. The mode used for a simulation is defined by setting the
# mode_obj.selected_mode index. The first is 0, the second 1 etc. EMerge will automatically bundle our data.

for i1, i2 in m.parameter_sweep(False, p1=(0,1), p2=(0,1)):
    p1.selected_mode = i1
    p2.selected_mode = i2
    data = m.mw.run_sweep()

# We can now visualize the mode fields in order. Please keep in mind that by default we look into the -X, -Y, -Z direction.

for i in range(2):
    for j in range(2):
        m.display.add_object(wg)
        m.display.add_portmode(p1, 30, k0=data.field[0].k0, mode_number=i)
        m.display.add_portmode(p2, 30, k0=data.field[0].k0, mode_number=j)
        m.display.show()

# We extract the gritted data if we want and verify that the coupling between the modes is low.

sgrid = data.scalar.grid

S21 = sgrid.S(2,1)
S21_11 = S21[0,0,:]
S21_21 = S21[1,0,:]
S21_12 = S21[0,1,:]
S21_22 = S21[1,1,:]

plot_sp(sgrid.freq[0,0,:], [S21_11, S21_12, S21_21, S21_22], labels=['T11','T12','T21','T22'])


import numpy as np
from ...material import Material
from ...const import Z0 as n0

PI = np.pi
TAU = 2*PI

def microstrip_z0(W: float, th: float, er: float):
    u = W/th
    fu = 6 + (TAU - 6)*np.exp(-((30.666/u)**(0.7528)))
    Z0 = n0/(TAU*np.sqrt(er))* np.log(fu/u + np.sqrt(1+(2/u)**2))
    return Z0


class PCBCalculator:

    def __init__(self, thickness: float, layers: np.ndarray, material: Material, unit: float):
        self.th = thickness
        self.layers = layers
        self.mat = material
        self.unit = unit

    def z0(self, Z0: float, layer: int = -1, ground_layer: int = 0, f0: float = 1e9):
        th = abs(self.layers[layer] - self.layers[ground_layer])*self.unit
        ws = np.geomspace(1e-6,1e-1,101)
        Z0ms = microstrip_z0(ws, th, self.mat.er.scalar(f0))
        return np.interp(Z0, Z0ms[::-1], ws[::-1])/self.unit
